#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>

#include "MNLClient.h"

/**
  * The port on which the MNLDaemon is listening.
  * (host byte order)
  */
uint16_t gMNLDaemonPort = 0;

/** The address of the UDP socket on which the daemon is listening.
  */
struct sockaddr_in gMNLDaemonAddr;

/** The socket to communicate with the daemon */
int gToDaemonSkt = 0;


/**
  * @return 0 if initialization failed. 1 if initialized succeeded.
  */
int MNLClientInit() {
  FILE* lFd = fopen(MNL_SOCK_FILE, "r");
  char lpLine [100];

  if (lFd == NULL) {
    printf("MNL: ClientInit: Unable to open MNLDaemon socket file %s.  Is the MNLDaemon running in the current directory? (%s)\n",
            MNL_SOCK_FILE,
            strerror(errno)
          );
    return 0;
  }
  
  if (fgets(lpLine, sizeof(lpLine), lFd) == NULL) {
    printf ("MNL: Init(): Unable to read MNLDaemon port information from %s\n",
          MNL_SOCK_FILE
        );
    fclose (lFd);
    return 0;
  } 
  gMNLDaemonPort = atoi(lpLine);
  if (gMNLDaemonPort == 0) {
    printf ("MNL: Init(): Unable to read MNLDaemon port information from %s\n",
          MNL_SOCK_FILE
        );
    fclose (lFd);
  }
 
  fclose (lFd);

  gMNLDaemonAddr.sin_family = AF_INET;
  gMNLDaemonAddr.sin_addr.s_addr = inet_addr("127.0.0.1");
  gMNLDaemonAddr.sin_port = htons(gMNLDaemonPort);
  memset(&(gMNLDaemonAddr.sin_zero), '\0', sizeof(gMNLDaemonAddr.sin_zero));

  gToDaemonSkt = socket(PF_INET, SOCK_DGRAM, 0);
  if (gToDaemonSkt < 0) { 
    printf (
          "MNL: ClientInit: Unable to open UDP socket to daemon: %s.\n",
          strerror (errno)
        );
    return 0;
  }
 
  {
    struct sockaddr_in lMyAddr;

    lMyAddr.sin_family = AF_INET;         
    lMyAddr.sin_port = 0;     
    lMyAddr.sin_addr.s_addr = INADDR_ANY; 
    memset(&(lMyAddr.sin_zero), '\0', sizeof(lMyAddr.sin_zero));
    
    if (bind(
              gToDaemonSkt,
              (struct sockaddr *)&lMyAddr, 
              sizeof(struct sockaddr_in)
            ) == -1) {

      printf ("MNL: ClientInit: Unable to bind UDP socket : %s\n", strerror (errno));
      return 0;
    }
  }

  printf ("MNL: Using MNLDaemon at port %d\n", gMNLDaemonPort);

  return 1;
}


/**
  * The first 6 parameters and return value have the same meaning as the standard sendto function.
  * @param ipError has an additional meaning : If some real error occurs while internally calling sendto,
  * this variable is set to 1.  Otherwise it is 0. Not that dropping a packet purposely is NOT an
  * error.  
  * @param iSeqNumStart The sequence number of the first byte of actual data contained in buf.
  * @param iDataLen The number of bytes of actual data contained in buf.
  * @param iWindowSize The current window size in bytes (if using sliding window, 0 otherwise).
  */
ssize_t MNL_sendto(int s, const void* buf, size_t len, int flags, const struct sockaddr *to, socklen_t tolen, 
                  int iSeqNumStart, int iDataLen, int iWindowSize, int *ipError) {

  ssize_t lRetVal = len;
  int lDropFlag = 0;
  static char *lpMsgBuf = NULL;
  int lNumSent = 0;
  
  SendMsg* lpMsg = NULL;
  int lMsgSize = sizeof(SendMsg) + len;

  if (gToDaemonSkt == 0) {
    printf ("MNL: sendto: Not initialized. Please call MNLClientInit() at the start of your program.\n");
    *ipError = 1;
    return -1;
  }
  
  if (len > MAX_PKT_LEN) {
    printf ("MNL: sendto(): Too large packet.  Maximum size is %d\n",
            MAX_PKT_LEN
          );
    *ipError = 1;
    return -1;
  }
  
  if (lpMsgBuf == NULL) {
    lpMsgBuf = (char *) malloc (sizeof(SendMsg) + MAX_PKT_LEN);
    if (lpMsgBuf == NULL) {
      printf ("MNL: sendto(): Unable to malloc packet for sending message to daemon:%s\n",
          strerror(errno)
        );
      *ipError = 1;
      return -1;
    }
  }
  
  *ipError = 0;

  lpMsg = (SendMsg *) lpMsgBuf;
  lpMsg->mPktLen = len;
  lpMsg->mToAddr = *( (struct sockaddr_in *) to);
  lpMsg->mToLen = tolen;
  lpMsg->mFlags = flags;
  lpMsg->mSeqNumStart = iSeqNumStart;
  lpMsg->mDataLen = iDataLen;
  lpMsg->mWindowSize = iWindowSize;
  memcpy(lpMsg->mpPktContents, buf, len);

  lNumSent = sendto(gToDaemonSkt,
                    lpMsgBuf,
                    lMsgSize,
                    flags,
                    (struct sockaddr *) &gMNLDaemonAddr,
                    sizeof (struct sockaddr_in)
                );
  if (lNumSent != lMsgSize) {
    *ipError = 1;
    return -1;
  }
                    

  return lRetVal;

}

